/*
 * Copyright © 2022 Soren Stoutner <soren@stoutner.com>.
 *
 * This file is part of Privacy Browser PC <https://www.stoutner.com/privacy-browser-pc>.
 *
 * Privacy Browser PC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Privacy Browser PC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Privacy Browser PC.  If not, see <http://www.gnu.org/licenses/>.
 */

// Application headers.
#include "BrowserView.h"
#include "MouseEventFilter.h"
#include "Settings.h"
#include "ui_BrowserView.h"
#include "helpers/DomainsDatabaseHelper.h"
#include "helpers/SearchEngineHelper.h"
#include "helpers/UserAgentHelper.h"
#include "interceptors/UrlRequestInterceptor.h"
#include "windows/BrowserWindow.h"

// Qt framework headers.
#include <QAction>
#include <QWebEngineProfile>

// Initialize the public static variables.
QString BrowserView::webEngineDefaultUserAgent = QStringLiteral("");

BrowserView::BrowserView(QWidget *parent) : QWidget(parent)
{
    // Instantiate the browser view UI.
    Ui::BrowserView browserViewUi;

    // Setup the UI.
    browserViewUi.setupUi(this);

    // Get handles for the views.
    webEngineViewPointer = browserViewUi.webEngineView;

    // Create an off-the-record profile (the default when no profile name is specified).
    webEngineProfilePointer = new QWebEngineProfile(QStringLiteral(""));

    // Create a WebEngine page.
    QWebEnginePage *webEnginePagePointer = new QWebEnginePage(webEngineProfilePointer);

    // Set the WebEngine page.
    webEngineViewPointer->setPage(webEnginePagePointer);

    // Get handles for the aspects of the WebEngine.
    webEngineHistoryPointer = webEnginePagePointer->history();
    webEngineSettingsPointer = webEngineViewPointer->settings();

    // Store a copy of the WebEngine default user agent.
    webEngineDefaultUserAgent = webEngineProfilePointer->httpUserAgent();

    // Update the URL line edit when the URL changes.
    connect(webEngineViewPointer, SIGNAL(urlChanged(const QUrl)), this, SLOT(updateUrl(const QUrl)));

    // Update the progress bar.
    connect(webEngineViewPointer, SIGNAL(loadStarted()), this, SLOT(loadStarted()));
    connect(webEngineViewPointer, SIGNAL(loadProgress(const int)), this, SLOT(loadProgress(const int)));
    connect(webEngineViewPointer, SIGNAL(loadFinished(const bool)), this, SLOT(loadFinished()));

    // Instantiate the mouse event filter pointer.
    MouseEventFilter *mouseEventFilterPointer = new MouseEventFilter(webEngineViewPointer);

    // Install the mouse event filter.
    qApp->installEventFilter(mouseEventFilterPointer);

    // Listen for hovered link URLs.
    connect(webEnginePagePointer, SIGNAL(linkHovered(const QString)), this, SLOT(pageLinkHovered(const QString)));

    // Instantiate the URL request interceptor.
    UrlRequestInterceptor *urlRequestInterceptorPointer = new UrlRequestInterceptor();

    // Set the URL request interceptor.
    webEngineProfilePointer->setUrlRequestInterceptor(urlRequestInterceptorPointer);

    // Reapply the domain settings when the host changes.
    connect(urlRequestInterceptorPointer, SIGNAL(applyDomainSettings(QString)), this, SLOT(applyDomainSettingsWithoutReloading(QString)));

    // Don't allow JavaScript to open windows.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptCanOpenWindows, false);

    // Allow keyboard navigation.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::SpatialNavigationEnabled, true);

    // Enable full screen support.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::FullScreenSupportEnabled, true);

    // Require user interaction to play media.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::PlaybackRequiresUserGesture, true);

    // Limit WebRTC to public IP addresses.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::WebRTCPublicInterfacesOnly, true);

    // Set the focus on the WebEngine view.
    webEngineViewPointer->setFocus();
}

void BrowserView::applyApplicationSettings()
{
    // Set the search engine URL.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(Settings::searchEngine());

    // Emit the update search engine actions signal.
    emit updateSearchEngineActions(Settings::searchEngine());
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettingsAndReload()
{
    // Apply the domain settings.  `true` reloads the website.
    applyDomainSettings(webEngineViewPointer->url().host(), true);
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettingsWithoutReloading(const QString &hostname)
{
    // Apply the domain settings  `false` does not reload the website.
    applyDomainSettings(hostname, false);
}

// Once <https://redmine.stoutner.com/issues/799> has been resolved this can be `const`.
void BrowserView::applyDomainSettings(const QString &hostname, const bool reloadWebsite)
{
    // Get the record for the hostname.
    QSqlQuery domainQuery = DomainsDatabaseHelper::getDomainQuery(hostname);

    // Check if the hostname has domain settings.
    if (domainQuery.isValid())  // The hostname has domain settings.
    {
        // Get the domain record.
        QSqlRecord domainRecord = domainQuery.record();

        // Set the JavaScript status.
        switch (domainRecord.field(DomainsDatabaseHelper::JAVASCRIPT).value().toInt())
        {
            case (DomainsDatabaseHelper::SYSTEM_DEFAULT):
            {
                // Set the default JavaScript status.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScript());

                break;
            }

            case (DomainsDatabaseHelper::DISABLED):
            {
                // Disable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, false);

                break;
            }

            case (DomainsDatabaseHelper::ENABLED):
            {
                // Enable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, true);

                break;
            }
        }

        // Set local storage.
        switch (domainRecord.field(DomainsDatabaseHelper::LOCAL_STORAGE).value().toInt())
        {
            case (DomainsDatabaseHelper::SYSTEM_DEFAULT):
            {
                // Set the default local storage status.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, Settings::localStorage());

                break;
            }

            case (DomainsDatabaseHelper::DISABLED):
            {
                // Disable local storage.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, false);

                break;
            }

            case (DomainsDatabaseHelper::ENABLED):
            {
                // Enable local storage.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, true);

                break;
            }
        }

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getResultingDomainSettingsUserAgent(domainRecord.field(DomainsDatabaseHelper::USER_AGENT).value().toString()));

        // Check if a custom zoom factor is set.
        if (domainRecord.field(DomainsDatabaseHelper::ZOOM_FACTOR).value().toInt())
        {
            // Store the current zoom factor.
            currentZoomFactor = domainRecord.field(DomainsDatabaseHelper::CUSTOM_ZOOM_FACTOR).value().toDouble();
        }
        else
        {
            // Reset the current zoom factor.
            currentZoomFactor = Settings::zoomFactor();
        }

        // Set the zoom factor.    The use of `currentZoomFactor` can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
        webEngineViewPointer->setZoomFactor(currentZoomFactor);

        // Apply the domain settings palette to the URL line edit.
        emit updateDomainSettingsIndicator(true, domainRecord.field(DomainsDatabaseHelper::DOMAIN_NAME).value().toString());
    }
    else  // The hostname does not have domain settings.
    {
        // Set the JavaScript status.
        webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScript());

        // Set local storage.
        webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, Settings::localStorage());

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromDatabaseName(Settings::userAgent()));

        // Store the current zoom factor.  This can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
        currentZoomFactor = Settings::zoomFactor();

        // Set the zoom factor.
        webEngineViewPointer->setZoomFactor(Settings::zoomFactor());

        // Apply the no domain settings palette to the URL line edit.
        emit updateDomainSettingsIndicator(false, QStringLiteral(""));
    }

    // Emit the update actions signals.
    emit updateJavaScriptAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));
    emit updateLocalStorageAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));
    emit updateUserAgentActions(webEngineProfilePointer->httpUserAgent());
    emit updateZoomFactorAction(webEngineViewPointer->zoomFactor());

    // Reload the website if requested.
    if (reloadWebsite)
    {
        webEngineViewPointer->reload();
    }
}

void BrowserView::applyOnTheFlySearchEngine(QAction *searchEngineActionPointer)
{
    // Store the search engine name.
    QString searchEngineName = searchEngineActionPointer->text();

    // Strip out any `&` characters.
    searchEngineName.remove('&');

    // Store the search engine string.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(searchEngineName);
}

void BrowserView::applyOnTheFlyUserAgent(QAction *userAgentActionPointer) const
{
    // Get the user agent name.
    QString userAgentName = userAgentActionPointer->text();

    // Strip out any `&` characters.
    userAgentName.remove('&');

    // Apply the user agent.
    webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromTranslatedName(userAgentName));

    // Reload the website.
    webEngineViewPointer->reload();
}

// This can be const once <https://redmine.stoutner.com/issues/799> has been resolved.
void BrowserView::applyOnTheFlyZoomFactor(const double &zoomFactor)
{
    // Update the current zoom factor.  This can be removed once <https://redmine.stoutner.com/issues/799> has been resolved.
    currentZoomFactor = zoomFactor;

    // Set the zoom factor.
    webEngineViewPointer->setZoomFactor(zoomFactor);
}

void BrowserView::back() const
{
    // Go back.
    webEngineViewPointer->back();
}

void BrowserView::forward() const
{
    // Go forward.
    webEngineViewPointer->forward();
}

void BrowserView::home() const
{
    // Load the homepage.
    webEngineViewPointer->load(QUrl::fromUserInput(Settings::homepage()));
}

void BrowserView::loadFinished() const
{
    // Hide the progress bar.
    emit hideProgressBar();
}

void BrowserView::loadInitialWebsite()
{
    // Apply the application settings.
    applyApplicationSettings();

    // Get the arguments.
    QStringList argumentsStringList = qApp->arguments();

    // Check to see if the arguments lists contains a URL.
    if (argumentsStringList.size() > 1)
    {
        // Load the URL from the arguments list.
        webEngineViewPointer->load(QUrl::fromUserInput(argumentsStringList.at(1)));
    }
    else
    {
        // Load the homepage.
        home();
    }
}

void BrowserView::loadProgress(const int &progress) const
{
    // Show the progress bar.
    emit showProgressBar(progress);
}

void BrowserView::loadStarted() const
{
    // Show the progress bar.
    emit showProgressBar(0);
}

void BrowserView::loadUrlFromLineEdit(QString url) const
{
    // Decide if the text is more likely to be a URL or a search.
    if (url.startsWith("file://"))  // The text is likely a file URL.
    {
        // Load the URL.
        webEngineViewPointer->load(QUrl::fromUserInput(url));
    }
    else if (url.contains("."))  // The text is likely a URL.
    {
        // Check if the URL does not start with a valid protocol.
        if (!url.startsWith("http"))
        {
            // Add `https://` to the beginning of the URL.
            url = "https://" + url;
        }

        // Load the URL.
        webEngineViewPointer->load(QUrl::fromUserInput(url));
    }
    else  // The text is likely a search.
    {
        // Load the search.
        webEngineViewPointer->load(QUrl::fromUserInput(searchEngineUrl + url));
    }
}

void BrowserView::pageLinkHovered(const QString &linkUrl) const
{
    // Emit a signal so that the browser window can update the status bar.
    emit linkHovered(linkUrl);
}

void BrowserView::refresh() const
{
    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleJavaScript() const
{
    // Toggle JavaScript.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, !webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));

    // Update the JavaScript icon.
    emit updateJavaScriptAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleLocalStorage() const
{
    // Toggle local storage.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::LocalStorageEnabled, !webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));

    // Update the local storage icon.
    emit updateLocalStorageAction(webEngineSettingsPointer->testAttribute(QWebEngineSettings::LocalStorageEnabled));

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::updateUrl(const QUrl &url) const
{
    // Update the URL line edit.
    emit updateUrlLineEdit(url);

    // Update the status of the forward and back buttons.
    emit updateBackAction(webEngineHistoryPointer->canGoBack());
    emit updateForwardAction(webEngineHistoryPointer->canGoForward());

    // Reapply the zoom factor.  This is a bug in QWebEngineView that resets the zoom with every load.  <https://redmine.stoutner.com/issues/799>
    webEngineViewPointer->setZoomFactor(currentZoomFactor);
}
